import unittest
from urlchecker_db import LogDB
import peewee
import urlchecker_dns
from urlchecker_dns import (
    expandDomainList,
    logDnsQuery,
    buildResolver,
    queryDNS,
    getReachableResolverList,
)
from urlchecker_status import logStatus
import mock


class MockAnswer(object):
    def __init__(self, address):
        self.address = address


class UrlCheckerDNSTestCase(unittest.TestCase):
    def setUp(self):
        self.db = LogDB(":memory:")
        self.db.createTables()

    ################################################
    # expandDomainList
    ################################################
    def test_expandDomainList(self):
        result = expandDomainList(["c", "b.a.a", "a.a.a", "a.a"])
        assert result == ["a.a", "a.a.a", "b.a.a", "c"]

    ################################################
    # logDnsQuery
    ################################################
    def test_logDnsQuery_insertFirst(self):
        domain = "http://example.org"
        resolver_ip = "127.0.0.1"
        rdtype = "foo"
        answer_list = ["4.3.2.1", "1.2.3.4"]
        status_id = logStatus(self.db, "foo")
        result = logDnsQuery(
            self.db, status_id, resolver_ip, domain, rdtype, answer_list
        )
        assert self.db.DnsChange.select().count() == 1
        assert self.db.DnsChange.get().resolver_ip == resolver_ip
        assert self.db.DnsChange.get().domain == domain
        assert self.db.DnsChange.get().rdtype == rdtype
        assert self.db.DnsChange.get().response == "1.2.3.4, 4.3.2.1"
        assert self.db.DnsChange.get().status_id == status_id
        assert result == status_id

    def test_logDnsQuery_insertOnlyOnePerStatusIdIPUrl(self):
        domain = "http://example.org"
        resolver_ip = "127.0.0.1"
        rdtype = "foo"
        answer_list = ["4.3.2.1", "1.2.3.4"]
        status_id = logStatus(self.db, "foo")
        logDnsQuery(
            self.db, status_id, resolver_ip, domain, rdtype, answer_list
        )
        answer_list.extend("127.0.0.1")
        try:
            logDnsQuery(
                self.db, status_id, resolver_ip, domain, rdtype, answer_list
            )
        except peewee.IntegrityError:
            assert self.db.DnsChange.select().count() == 1
            assert self.db.DnsChange.get().status_id == status_id
        else:
            raise NotImplementedError("Expected IntegrityError")

    def test_logDnsQuery_skipIdenticalPreviousValues(self):
        domain = "http://example.org"
        resolver_ip = "127.0.0.1"
        rdtype = "foo"
        answer_list = ["4.3.2.1", "1.2.3.4"]
        status_id = logStatus(self.db, "foo")
        result = logDnsQuery(
            self.db, status_id, resolver_ip, domain, rdtype, answer_list
        )
        status_id_2 = logStatus(self.db, "foo")
        result_2 = logDnsQuery(
            self.db, status_id_2, resolver_ip, domain, rdtype, answer_list
        )

        assert self.db.DnsChange.select().count() == 1
        assert self.db.DnsChange.get().resolver_ip == resolver_ip
        assert self.db.DnsChange.get().domain == domain
        assert self.db.DnsChange.get().rdtype == rdtype
        assert self.db.DnsChange.get().response == "1.2.3.4, 4.3.2.1"
        assert self.db.DnsChange.get().status_id == status_id
        assert result == status_id
        assert result == result_2

    def test_logDnsQuery_insertWhenDifferentAnswer(self):
        domain = "http://example.org"
        resolver_ip = "127.0.0.1"
        rdtype = "foo"
        answer_list = ["4.3.2.1", "1.2.3.4"]
        status_id = logStatus(self.db, "foo")
        result = logDnsQuery(
            self.db, status_id, resolver_ip, domain, rdtype, answer_list
        )
        answer_list_2 = ["4.3.2.1", "1.2.3.4", "0.0.0.0"]
        status_id_2 = logStatus(self.db, "foo")
        result_2 = logDnsQuery(
            self.db, status_id_2, resolver_ip, domain, rdtype, answer_list_2
        )

        assert result_2 != result
        assert self.db.DnsChange.select().count() == 2

        assert (
            self.db.DnsChange.get(
                self.db.DnsChange.status == status_id
            ).resolver_ip
            == resolver_ip
        )
        assert (
            self.db.DnsChange.get(self.db.DnsChange.status == status_id).domain
            == domain
        )
        assert (
            self.db.DnsChange.get(self.db.DnsChange.status == status_id).rdtype
            == rdtype
        )
        assert (
            self.db.DnsChange.get(
                self.db.DnsChange.status == status_id
            ).response
            == "1.2.3.4, 4.3.2.1"
        )
        assert (
            self.db.DnsChange.get(
                self.db.DnsChange.status == status_id
            ).status_id
            == status_id
        )

        assert (
            self.db.DnsChange.get(
                self.db.DnsChange.status == status_id_2
            ).resolver_ip
            == resolver_ip
        )
        assert (
            self.db.DnsChange.get(
                self.db.DnsChange.status == status_id_2
            ).domain
            == domain
        )
        assert (
            self.db.DnsChange.get(
                self.db.DnsChange.status == status_id_2
            ).rdtype
            == rdtype
        )
        assert (
            self.db.DnsChange.get(
                self.db.DnsChange.status == status_id_2
            ).response
            == "0.0.0.0, 1.2.3.4, 4.3.2.1"
        )
        assert (
            self.db.DnsChange.get(
                self.db.DnsChange.status == status_id_2
            ).status_id
            == status_id_2
        )

    def test_logDnsQuery_insertDifferentUrl(self):
        domain = "http://example.org"
        domain_2 = domain + "."
        resolver_ip = "127.0.0.1"
        resolver_ip_2 = resolver_ip + "1"
        rdtype = "foo"
        rdtype_2 = rdtype + "bar"
        answer_list = ["4.3.2.1", "1.2.3.4"]
        status_id = logStatus(self.db, "foo")
        logDnsQuery(
            self.db, status_id, resolver_ip, domain, rdtype, answer_list
        )
        logDnsQuery(
            self.db, status_id, resolver_ip_2, domain, rdtype, answer_list
        )
        logDnsQuery(
            self.db, status_id, resolver_ip, domain_2, rdtype, answer_list
        )
        logDnsQuery(
            self.db, status_id, resolver_ip, domain, rdtype_2, answer_list
        )
        logDnsQuery(
            self.db, status_id, resolver_ip_2, domain_2, rdtype, answer_list
        )
        logDnsQuery(
            self.db, status_id, resolver_ip_2, domain, rdtype_2, answer_list
        )
        logDnsQuery(
            self.db, status_id, resolver_ip, domain_2, rdtype_2, answer_list
        )
        logDnsQuery(
            self.db, status_id, resolver_ip_2, domain_2, rdtype_2, answer_list
        )

        assert self.db.DnsChange.select().count() == 8

    ################################################
    # buildResolver
    ################################################
    def test_buildResolver_default(self):
        resolver = buildResolver("127.0.0.1")
        assert resolver.nameservers == ["127.0.0.1"]
        assert resolver.timeout == 2
        assert resolver.lifetime == 2
        assert resolver.edns == -1

    ################################################
    # queryDNS
    ################################################
    def test_queryDNS_default(self):
        resolver_ip = "127.0.0.1"
        domain = "example.org"
        rdtype = "A"
        status_id = logStatus(self.db, "foo")

        with mock.patch(
            "urlchecker_dns.dns.resolver.Resolver.query"
        ) as mock_query:
            mock_query.return_value = [
                MockAnswer("4.3.2.1"),
                MockAnswer("1.2.3.4"),
            ]
            result = queryDNS(self.db, status_id, resolver_ip, domain, rdtype)

            assert mock_query.call_count == 1
            mock_query.assert_called_with(
                domain, rdtype, raise_on_no_answer=False
            )

        assert self.db.DnsChange.select().count() == 1
        assert self.db.DnsChange.get().resolver_ip == resolver_ip
        assert self.db.DnsChange.get().domain == domain
        assert self.db.DnsChange.get().rdtype == rdtype
        assert self.db.DnsChange.get().response == "1.2.3.4, 4.3.2.1"
        assert self.db.DnsChange.get().status_id == status_id
        assert result == ["1.2.3.4", "4.3.2.1"]

    def test_queryDNS_rejectRdtype(self):
        resolver_ip = "127.0.0.1"
        domain = "example.org"
        rdtype = "B"
        status_id = logStatus(self.db, "foo")

        try:
            queryDNS(self.db, status_id, resolver_ip, domain, rdtype)
        except AssertionError:
            assert self.db.DnsChange.select().count() == 0
        else:
            raise NotImplementedError("Expected AssertionError")

    def test_queryDNS_nxdomain(self):
        resolver_ip = "127.0.0.1"
        domain = "example.org"
        rdtype = "A"
        status_id = logStatus(self.db, "foo")

        with mock.patch(
            "urlchecker_dns.dns.resolver.Resolver.query"
        ) as mock_query:

            def sideEffect(*args, **kw):
                raise urlchecker_dns.dns.resolver.NXDOMAIN()

            mock_query.side_effect = sideEffect
            result = queryDNS(self.db, status_id, resolver_ip, domain, rdtype)

            assert mock_query.call_count == 1
            mock_query.assert_called_with(
                domain, rdtype, raise_on_no_answer=False
            )

        assert self.db.DnsChange.select().count() == 1
        assert self.db.DnsChange.get().resolver_ip == resolver_ip
        assert self.db.DnsChange.get().domain == domain
        assert self.db.DnsChange.get().rdtype == rdtype
        assert self.db.DnsChange.get().response == ""
        assert self.db.DnsChange.get().status_id == status_id
        assert result == []

    def test_queryDNS_NoAnswer(self):
        resolver_ip = "127.0.0.1"
        domain = "example.org"
        rdtype = "A"
        status_id = logStatus(self.db, "foo")

        with mock.patch(
            "urlchecker_dns.dns.resolver.Resolver.query"
        ) as mock_query:

            def sideEffect(*args, **kw):
                raise urlchecker_dns.dns.resolver.NoAnswer()

            mock_query.side_effect = sideEffect
            result = queryDNS(self.db, status_id, resolver_ip, domain, rdtype)

            assert mock_query.call_count == 1
            mock_query.assert_called_with(
                domain, rdtype, raise_on_no_answer=False
            )

        assert self.db.DnsChange.select().count() == 1
        assert self.db.DnsChange.get().resolver_ip == resolver_ip
        assert self.db.DnsChange.get().domain == domain
        assert self.db.DnsChange.get().rdtype == rdtype
        assert self.db.DnsChange.get().response == ""
        assert self.db.DnsChange.get().status_id == status_id
        assert result == []

    def test_queryDNS_timeout(self):
        resolver_ip = "127.0.0.1"
        domain = "example.org"
        rdtype = "A"
        status_id = logStatus(self.db, "foo")

        with mock.patch(
            "urlchecker_dns.dns.resolver.Resolver.query"
        ) as mock_query:

            def sideEffect(*args, **kw):
                raise urlchecker_dns.dns.exception.Timeout()

            mock_query.side_effect = sideEffect
            result = queryDNS(self.db, status_id, resolver_ip, domain, rdtype)

            assert mock_query.call_count == 1
            mock_query.assert_called_with(
                domain, rdtype, raise_on_no_answer=False
            )

        assert self.db.DnsChange.select().count() == 1
        assert self.db.DnsChange.get().resolver_ip == resolver_ip
        assert self.db.DnsChange.get().domain == domain
        assert self.db.DnsChange.get().rdtype == rdtype
        assert self.db.DnsChange.get().response == ""
        assert self.db.DnsChange.get().status_id == status_id
        assert result == []

    def test_queryDNS_nonameservers(self):
        resolver_ip = "127.0.0.1"
        domain = "example.org"
        rdtype = "A"
        status_id = logStatus(self.db, "foo")

        with mock.patch(
            "urlchecker_dns.dns.resolver.Resolver.query"
        ) as mock_query:

            def sideEffect(*args, **kw):
                raise urlchecker_dns.dns.resolver.NoNameservers()

            mock_query.side_effect = sideEffect
            result = queryDNS(self.db, status_id, resolver_ip, domain, rdtype)

            assert mock_query.call_count == 1
            mock_query.assert_called_with(
                domain, rdtype, raise_on_no_answer=False
            )

        assert self.db.DnsChange.select().count() == 1
        assert self.db.DnsChange.get().resolver_ip == resolver_ip
        assert self.db.DnsChange.get().domain == domain
        assert self.db.DnsChange.get().rdtype == rdtype
        assert self.db.DnsChange.get().response == ""
        assert self.db.DnsChange.get().status_id == status_id
        assert result == []

    ################################################
    # getReachableResolverList
    ################################################
    def test_getReachableResolverList_open(self):
        resolver_ip = "127.0.0.1"
        domain = "example.org"
        rdtype = "A"
        status_id = logStatus(self.db, "foo")

        with mock.patch(
            "urlchecker_dns.dns.resolver.Resolver.query"
        ) as mock_query:
            mock_query.return_value = [
                MockAnswer("4.3.2.1"),
                MockAnswer("1.2.3.4"),
            ]
            result = getReachableResolverList(
                self.db, status_id, [resolver_ip]
            )

            assert mock_query.call_count == 1
            mock_query.assert_called_with(
                domain, rdtype, raise_on_no_answer=False
            )

        assert self.db.DnsChange.select().count() == 1
        assert self.db.DnsChange.get().resolver_ip == resolver_ip
        assert self.db.DnsChange.get().domain == domain
        assert self.db.DnsChange.get().rdtype == rdtype
        assert self.db.DnsChange.get().response == "1.2.3.4, 4.3.2.1"
        assert self.db.DnsChange.get().status_id == status_id
        assert result == ["127.0.0.1"]

        assert self.db.NetworkChange.select().count() == 1
        assert self.db.NetworkChange.get().ip == resolver_ip
        assert self.db.NetworkChange.get().port == 53
        assert self.db.NetworkChange.get().transport == "UDP"
        assert self.db.NetworkChange.get().state == "open"
        assert self.db.NetworkChange.get().status_id == status_id

    def test_getReachableResolverList_closed(self):
        resolver_ip = "127.0.0.1"
        domain = "example.org"
        rdtype = "A"
        status_id = logStatus(self.db, "foo")

        with mock.patch(
            "urlchecker_dns.dns.resolver.Resolver.query"
        ) as mock_query:

            def sideEffect(*args, **kw):
                raise urlchecker_dns.dns.exception.Timeout()

            mock_query.side_effect = sideEffect

            result = getReachableResolverList(
                self.db, status_id, [resolver_ip]
            )

            assert mock_query.call_count == 1
            mock_query.assert_called_with(
                domain, rdtype, raise_on_no_answer=False
            )

        assert self.db.DnsChange.select().count() == 1
        assert self.db.DnsChange.get().resolver_ip == resolver_ip
        assert self.db.DnsChange.get().domain == domain
        assert self.db.DnsChange.get().rdtype == rdtype
        assert self.db.DnsChange.get().response == ""
        assert self.db.DnsChange.get().status_id == status_id
        assert result == []

        assert self.db.NetworkChange.select().count() == 1
        assert self.db.NetworkChange.get().ip == resolver_ip
        assert self.db.NetworkChange.get().port == 53
        assert self.db.NetworkChange.get().transport == "UDP"
        assert self.db.NetworkChange.get().state == "closed"
        assert self.db.NetworkChange.get().status_id == status_id

    def test_getReachableResolverList_noiplist(self):
        resolver_ip = "127.0.0.1"
        domain = "example.org"
        rdtype = "A"
        status_id = logStatus(self.db, "foo")

        resolver = urlchecker_dns.dns.resolver.Resolver(configure=False)
        resolver.nameservers.append(resolver_ip)

        with mock.patch(
            "urlchecker_dns.get_default_resolver"
        ) as mock_get_default_resolver:
            mock_get_default_resolver.return_value = resolver

            with mock.patch(
                "urlchecker_dns.dns.resolver.Resolver.query"
            ) as mock_query:
                mock_query.return_value = [
                    MockAnswer("4.3.2.1"),
                    MockAnswer("1.2.3.4"),
                ]
                result = getReachableResolverList(self.db, status_id, [])

                assert mock_get_default_resolver.call_count == 1

                assert mock_query.call_count == 1
                mock_query.assert_called_with(
                    domain, rdtype, raise_on_no_answer=False
                )

        assert self.db.DnsChange.select().count() == 1
        assert self.db.DnsChange.get().resolver_ip == resolver_ip
        assert self.db.DnsChange.get().domain == domain
        assert self.db.DnsChange.get().rdtype == rdtype
        assert self.db.DnsChange.get().response == "1.2.3.4, 4.3.2.1"
        assert self.db.DnsChange.get().status_id == status_id
        assert result == ["127.0.0.1"]

        assert self.db.NetworkChange.select().count() == 1
        assert self.db.NetworkChange.get().ip == resolver_ip
        assert self.db.NetworkChange.get().port == 53
        assert self.db.NetworkChange.get().transport == "UDP"
        assert self.db.NetworkChange.get().state == "open"
        assert self.db.NetworkChange.get().status_id == status_id


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(UrlCheckerDNSTestCase))
    return suite


if __name__ == "__main__":
    unittest.main(defaultTest="suite")
