import time
from urlchecker_db import LogDB
from urlchecker_configuration import createConfiguration, logConfiguration
from urlchecker_platform import logPlatform
from urlchecker_status import logStatus
from urlchecker_dns import (
    getReachableResolverList,
    expandDomainList,
    getServerIpDict,
)
from urlchecker_http import getUrlHostname, checkHttpStatus
from urlchecker_network import isTcpPortOpen


__version__ = "0.0.3"


class BotError(Exception):
    pass


class WebBot:
    def __init__(self, **kw):
        self.config = createConfiguration(**kw)

    def initDB(self, sqlite_path):
        self._db = LogDB(sqlite_path)
        self._db.createTables()

    def iterateLoop(self):
        status_id = logStatus(self._db, "loop")
        logPlatform(self._db, __version__, status_id)

        # Calculate the resolver list
        resolver_ip_list = getReachableResolverList(
            self._db, status_id, self.config["DNS"].split()
        )
        if not resolver_ip_list:
            return

        # Calculate the full list of domain to check
        domain_list = self.config["DOMAIN"].split()

        # Extract the list of URL domains
        url_list = self.config["URL"].split()
        for url in url_list:
            domain_list.append(getUrlHostname(url))
        domain_list = list(set(domain_list))

        # Expand with all parent domains
        domain_list = expandDomainList(domain_list)

        # Get the list of server to check
        # XXX Check DNS expiration
        server_ip_dict = getServerIpDict(
            self._db, status_id, resolver_ip_list, domain_list, "A"
        )

        # Check TCP port for the list of IP found
        # XXX For now, check http/https only
        server_ip_list = [x for x in server_ip_dict.keys()]
        url_dict = {}
        for server_ip in server_ip_list:
            # XXX Check SSL certificate expiration
            for port, protocol in [(80, "http"), (443, "https")]:
                if isTcpPortOpen(self._db, server_ip, port, status_id):
                    for hostname in server_ip_dict[server_ip]:
                        url = "%s://%s" % (protocol, hostname)
                        if url not in url_dict:
                            url_dict[url] = []
                        url_dict[url].append(server_ip)

        # XXX put back orignal url list

        # Check HTTP Status
        for url in url_dict:
            for ip in url_dict[url]:
                checkHttpStatus(self._db, status_id, url, ip, __version__)
                # XXX Check location header and check new url recursively
                # XXX Parse HTML, fetch found link, css, js, image
                # XXX Check HTTP Cache

    def stop(self):
        self._running = False
        logStatus(self._db, "stop")
        if hasattr(self, "_db"):
            self._db.close()

    def run(self):
        self.initDB(self.config["SQLITE"])
        status_id = logStatus(self._db, "start")
        logConfiguration(self._db, status_id, self.config)

        self._running = True
        try:
            while self._running:
                self.iterateLoop()
                interval = int(self.config.get("INTERVAL"))
                if interval < 0:
                    self.stop()
                else:
                    time.sleep(interval)
        except KeyboardInterrupt:
            self.stop()
        except:
            # XXX Put traceback in the log?
            self.stop()
            logStatus(self._db, "error")
            raise


def create_bot(**kw):
    return WebBot(**kw)
