import peewee
from playhouse.migrate import migrate, SqliteMigrator
from playhouse.sqlite_ext import SqliteExtDatabase


class LogDB:
    def __init__(self, sqlite_path):
        self._db = SqliteExtDatabase(
            sqlite_path, pragmas=(("journal_mode", "WAL"), ("foreign_keys", 1))
        )
        self._db.connect()

        class BaseModel(peewee.Model):
            class Meta:
                database = self._db

        # This store the start, stop, loop time of the bot
        # All other tables point to it to be able to group some info
        class Status(BaseModel):
            text = peewee.TextField(index=True)
            timestamp = peewee.TimestampField(
                index=True, constraints=[peewee.SQL("DEFAULT now")]
            )

        # Store the configuration modification
        class ConfigurationChange(BaseModel):
            status = peewee.ForeignKeyField(Status)
            parameter = peewee.TextField(index=True)
            value = peewee.TextField()

            class Meta:
                primary_key = peewee.CompositeKey("status", "parameter")

        # Store the configuration modification
        class PlatformChange(BaseModel):
            status = peewee.ForeignKeyField(Status)
            parameter = peewee.TextField(index=True)
            value = peewee.TextField()

            class Meta:
                primary_key = peewee.CompositeKey("status", "parameter")

        # Store remote network status
        class NetworkChange(BaseModel):
            status = peewee.ForeignKeyField(Status)
            ip = peewee.TextField()
            transport = peewee.TextField()
            port = peewee.IntegerField()
            state = peewee.TextField()

            class Meta:
                primary_key = peewee.CompositeKey(
                    "status", "ip", "transport", "port"
                )

        class DnsChange(BaseModel):
            status = peewee.ForeignKeyField(Status)
            resolver_ip = peewee.TextField()
            domain = peewee.TextField()
            rdtype = peewee.TextField()
            response = peewee.TextField()

            class Meta:
                primary_key = peewee.CompositeKey(
                    "status", "resolver_ip", "domain", "rdtype"
                )

        class HttpCodeChange(BaseModel):
            status = peewee.ForeignKeyField(Status)
            ip = peewee.TextField()
            url = peewee.TextField()
            status_code = peewee.IntegerField()

            class Meta:
                primary_key = peewee.CompositeKey("status", "ip", "url")

        self.Status = Status
        self.ConfigurationChange = ConfigurationChange
        self.PlatformChange = PlatformChange
        self.NetworkChange = NetworkChange
        self.DnsChange = DnsChange
        self.HttpCodeChange = HttpCodeChange

    def createTables(self):
        # http://www.sqlite.org/pragma.html#pragma_user_version
        db_version = self._db.pragma("user_version")
        expected_version = 1
        if db_version == 0:
            with self._db.transaction():
                self._db.create_tables(
                    [
                        self.Status,
                        self.ConfigurationChange,
                        self.HttpCodeChange,
                        self.NetworkChange,
                        self.PlatformChange,
                        self.DnsChange,
                    ]
                )
                self._db.pragma("user_version", expected_version)
        elif db_version != expected_version:
            # migrator = SqliteMigrator(self._db)
            migration_list = []
            sql_query_list = []

            if migration_list or sql_query_list:
                with self._db.transaction():
                    if migration_list:
                        migrate(*migration_list)
                    if sql_query_list:
                        for sql_query in sql_query_list:
                            self._db.execute_sql(
                                sql_query, require_commit=False
                            )
                    self._db.pragma("user_version", expected_version)

    def close(self):
        self._db.close()
