from dns.resolver import get_default_resolver
import dns
from urlchecker_network import logNetwork

URL_TO_CHECK = "example.org"
TIMEOUT = 2


def logDnsQuery(db, status_id, resolver_ip, domain_text, rdtype, answer_list):
    answer_list.sort()
    response = ", ".join(answer_list)

    with db._db.atomic():
        try:
            # Check previous parameter value
            previous_entry = (
                db.DnsChange.select()
                .where(
                    db.DnsChange.resolver_ip == resolver_ip,
                    db.DnsChange.domain == domain_text,
                    db.DnsChange.rdtype == rdtype,
                )
                .order_by(db.DnsChange.status.desc())
                .get()
            )
        except db.DnsChange.DoesNotExist:
            previous_entry = None

        if (previous_entry is None) or (previous_entry.response != response):
            previous_entry = db.DnsChange.create(
                resolver_ip=resolver_ip,
                domain=domain_text,
                rdtype=rdtype,
                response=response,
                status=status_id,
            )

    return previous_entry.status_id


def buildResolver(resolver_ip):
    resolver = dns.resolver.Resolver(configure=False)
    resolver.nameservers.append(resolver_ip)
    resolver.timeout = TIMEOUT
    resolver.lifetime = TIMEOUT
    resolver.edns = -1
    return resolver


def queryDNS(db, status_id, resolver_ip, domain_text, rdtype):
    # only A (and AAAA) has address property
    assert rdtype == "A"

    resolver = buildResolver(resolver_ip)
    try:
        answer_list = [
            x.address
            for x in resolver.query(
                domain_text, rdtype, raise_on_no_answer=False
            )
        ]
    except (
        dns.resolver.NXDOMAIN,
        dns.resolver.NoAnswer,
        dns.exception.Timeout,
        dns.resolver.NoNameservers,
    ):
        answer_list = []

    logDnsQuery(db, status_id, resolver_ip, domain_text, rdtype, answer_list)
    return answer_list


def getReachableResolverList(db, status_id, resolver_ip_list):
    # Create a list of resolver object
    if len(resolver_ip_list) == 0:
        resolver_ip_list = get_default_resolver().nameservers

    result_ip_list = []
    # Check the DNS server availability once
    # to prevent using it later if it is down
    for resolver_ip in resolver_ip_list:
        resolver_state = "open"
        answer_list = queryDNS(db, status_id, resolver_ip, URL_TO_CHECK, "A")

        if len(answer_list) == 0:
            # We expect a valid response
            # Drop the DNS server...
            resolver_state = "closed"
        else:
            resolver_state = "open"
            result_ip_list.append(resolver_ip)
        logNetwork(db, resolver_ip, "UDP", 53, resolver_state, status_id)

    return result_ip_list


def expandDomainList(domain_list):
    for domain_text in domain_list:

        dns_name = dns.name.from_text(domain_text)
        if (len(dns_name.labels) - 1) > 2:
            domain_list.append(dns_name.parent().to_text(omit_final_dot=True))

    domain_list = list(set(domain_list))
    domain_list.sort()
    return domain_list


def getServerIpDict(db, status_id, resolver_ip_list, domain_list, rdtype):
    server_ip_dict = {}
    for domain_text in domain_list:
        for resolver_ip in resolver_ip_list:
            answer_list = queryDNS(
                db, status_id, resolver_ip, domain_text, rdtype
            )
            for address in answer_list:
                if address not in server_ip_dict:
                    server_ip_dict[address] = []
                if domain_text not in server_ip_dict[address]:
                    # Do not duplicate the domain
                    server_ip_dict[address].append(domain_text)
    return server_ip_dict
